@@ -15,7 +15,8 @@
 # include <config.h>
 #endif
 #include "gif.h"
-#include <stdlib.h>
+#include <string.h>
+#include <stdarg.h>
 #ifdef __cplusplus
 extern "C" {
 #endif
@@ -60,7 +61,8 @@ Gif_NewImage(void)
   gfi->compressed_len = 0;
   gfi->compressed = 0;
   gfi->free_compressed = 0;
-  gfi->userdata = 0;
+  gfi->user_data = 0;
+  gfi->free_user_data = 0;
   gfi->refcount = 0;
   return gfi;
 }
@@ -244,11 +246,12 @@ Gif_CalculateScreenSize(Gif_Stream *gfs, int force)
   
   for (i = 0; i < gfs->nimages; i++) {
     Gif_Image *gfi = gfs->images[i];
-    if (gfi->left != 0 || gfi->top != 0) continue;
-    if (screen_width < gfi->width)
-      screen_width = gfi->width;
-    if (screen_height < gfi->height)
-      screen_height = gfi->height;
+    /* 17.Dec.1999 - I find this old behavior annoying. */
+    /* if (gfi->left != 0 || gfi->top != 0) continue; */
+    if (screen_width < gfi->left + gfi->width)
+      screen_width = gfi->left + gfi->width;
+    if (screen_height < gfi->top + gfi->height)
+      screen_height = gfi->top + gfi->height;
   }
   
   /* Only use the default 640x480 screen size if we are being forced to create
@@ -376,7 +379,6 @@ Gif_CopyImage(Gif_Image *src)
     dest->compressed_len = src->compressed_len;
   }
   
-  dest->userdata = src->userdata;
   return dest;
   
  failure:
@@ -446,6 +448,8 @@ Gif_DeleteImage(Gif_Image *gfi)
   Gif_DeleteArray(gfi->img);
   if (gfi->compressed && gfi->free_compressed)
     (*gfi->free_compressed)((void *)gfi->compressed);
+  if (gfi->user_data && gfi->free_user_data)
+    (*gfi->free_user_data)(gfi->user_data);
   Gif_Delete(gfi);
 }
 
@@ -727,22 +731,16 @@ Gif_CreateUncompressedImage(Gif_Image *gfi)
 }
 
 
-#ifdef GIF_DEBUGGING
-#include <stdarg.h>
-
 void
 Gif_Debug(char *x, ...)
 {
   va_list val;
   va_start(val, x);
   vfprintf(stderr, x, val);
-  fputc('\n', stderr);
+  fputc(' ', stderr);
   va_end(val);
 }
 
-#endif
-
-
 #ifdef __cplusplus
 }
 #endif
