@@ -1,5 +1,5 @@
 /* gifsicle.c - gifsicle's main loop.
-   Copyright (C) 1997-9 Eddie Kohler, eddietwo@lcs.mit.edu
+   Copyright (C) 1997-2000 Eddie Kohler, eddietwo@lcs.mit.edu
    This file is part of gifsicle.
 
    Gifsicle is free software. It is distributed under the GNU Public License,
@@ -52,10 +52,7 @@ static Gt_ColorTransform *output_transforms;
 static int mode = BLANK_MODE;
 static int nested_mode = 0;
 
-static FILE *infoing = 0;
-static int colormap_infoing = 0;
-static int extension_infoing = 0;
-static int outputing = 1;
+static int infoing = 0;
 int verbosing = 0;
 
 
@@ -172,6 +169,8 @@ static const char *output_option_types[] = {
 #define COLOR_TRANSFORM_OPT	353
 #define RESIZE_OPT		354
 #define SCALE_OPT		355
+#define NO_WARNINGS_OPT		356
+#define WARNINGS_OPT		357
 
 #define LOOP_TYPE		(Clp_MaxDefaultType + 1)
 #define DISPOSAL_TYPE		(Clp_MaxDefaultType + 2)
@@ -212,7 +211,7 @@ Clp_Option options[] = {
   { "explode", 'e', 'e', 0, 0 },
   { "explode-by-name", 'E', 'E', 0, 0 },
   { "extension", 0, EXTENSION_OPT, Clp_ArgString, 0 },
-  { "no-extensions", 0, NO_EXTENSIONS_OPT, 0, 0 },
+  { "no-extensions", 'x', NO_EXTENSIONS_OPT, 0, 0 },
   { "extension-info", 0, EXTENSION_INFO_OPT, 0, Clp_Negate },
   
   { "flip-horizontal", 0, FLIP_HORIZ_OPT, 0, Clp_Negate },
@@ -275,6 +274,9 @@ Clp_Option options[] = {
   { "verbose", 'v', VERBOSE_OPT, 0, Clp_Negate },
   { "version", 0, VERSION_OPT, 0, 0 },
   
+  { 0, 'w', NO_WARNINGS_OPT, 0, Clp_Negate },
+  { "warnings", 0, WARNINGS_OPT, 0, Clp_Negate },
+  
   { "xinfo", 0, EXTENSION_INFO_OPT, 0, Clp_Negate },
   
 };
@@ -397,6 +399,47 @@ show_frame(int imagenumber, int usename)
  * input a stream
  **/
 
+static int gifread_error_count;
+
+static void
+gifread_error(const char *message, int which_image, void *thunk)
+{
+  static int last_which_image = 0;
+  static const char *last_message = 0;
+  static int different_error_count = 0;
+  static int same_error_count = 0;
+  const char *filename = (const char *)thunk;
+  
+  if (gifread_error_count == 0) {
+    last_which_image = -1;
+    different_error_count = 0;
+  }
+  
+  gifread_error_count++;
+  if (last_message && different_error_count <= 10
+      && (message != last_message || last_which_image != which_image)) {
+    if (same_error_count == 1)
+      error("  %s", last_message);
+    else if (same_error_count > 0)
+      error("  %s (%d times)", last_message, same_error_count);
+    same_error_count = 0;
+  }
+
+  if (message != last_message)
+    different_error_count++;
+  
+  same_error_count++;
+  last_message = message;
+  if (last_which_image != which_image && different_error_count <= 10
+      && message) {
+    error("Error while reading `%s' frame #%d:", filename, which_image);
+    last_which_image = which_image;
+  }
+  
+  if (different_error_count == 11 && message)
+    error("(more errors while reading `%s')", filename);
+}
+
 void
 input_stream(char *name)
 {
@@ -440,8 +483,10 @@ input_stream(char *name)
   ungetc(i, f);
   
   if (verbosing) verbose_open('<', name);
-  gfs = Gif_FullReadFile(f, read_flags);
+  gifread_error_count = 0;
+  gfs = Gif_FullReadFile(f, read_flags, gifread_error, (void *)name);
   fclose(f);
+  gifread_error(0, -1, (void *)name); /* print out last error message */
   
   if (!gfs || (Gif_ImageCount(gfs) == 0 && gfs->errors > 0)) {
     error("%s: not a GIF", name);
@@ -451,10 +496,6 @@ input_stream(char *name)
   }
   
   input = gfs;
-  gfs->userflags = 97; /* to indicate no output done */
-  
-  if (gfs->errors)
-    error("%s is partially corrupted", name);
   
   /* Processing when we've got a new input frame */
   if (mode == BLANK_MODE)
@@ -466,13 +507,13 @@ input_stream(char *name)
     if (mode == BATCHING)
       active_output_data.output_name = input_name;
     else if (mode == EXPLODING) {
-      /* Explode into current directory. Use name instead of input_name: it's
-         #stdin# if required. */
-      char *slash = strrchr(name, PATHNAME_SEPARATOR);
+      /* Explode into current directory. */
+      char *explode_name = (input_name ? input_name : "#stdin#");
+      char *slash = strrchr(explode_name, PATHNAME_SEPARATOR);
       if (slash)
 	active_output_data.output_name = slash + 1;
       else
-	active_output_data.output_name = name;
+	active_output_data.output_name = explode_name;
     }
   }
   
@@ -502,9 +543,12 @@ input_stream(char *name)
     def_frame.comment = 0;
   if (!CHANGED(saved_next_frame, CH_EXTENSION))
     def_frame.extensions = 0;
+  def_frame.input_filename = input_name;
   
   old_def_frame = def_frame;
   first_input_frame = frames->count;
+  if (gfs->nimages > 1)
+    def_frame.position_is_offset = 1;
   for (i = 0; i < gfs->nimages; i++)
     add_frame(frames, -1, gfs, gfs->images[i]);
   def_frame = old_def_frame;
@@ -514,15 +558,13 @@ input_stream(char *name)
       static int context = 0;
       warning("`%s' is too complex to unoptimize", name);
       if (!context) {
-	warning("(The reason was local color tables or complex transparency.");
-	warning("Try running the GIF through `gifsicle --colors=255' first.)");
+	warncontext("(The reason was local color tables or complex transparency.");
+	warncontext("Try running the GIF through `gifsicle --colors=255' first.)");
       }
       context = 1;
     }
   
   apply_color_transforms(input_transforms, gfs);
-  if (infoing)
-    stream_info(infoing, gfs, name, colormap_infoing, extension_infoing);
   gfs->refcount++;
 }
 
@@ -532,15 +574,15 @@ input_done(void)
   if (!input) return;
   
   if (verbosing) verbose_close('>');
-  if (infoing) {
+  /*if (infoing) {
     int i;
-    if (input->userflags == 97)	/* no stream info produced yet */
+    if (input->userflags == 97)
       stream_info(infoing, input, input_name,
 		  colormap_infoing, extension_infoing);
     for (i = first_input_frame; i < frames->count; i++)
       if (FRAME(frames, i).stream == input && FRAME(frames, i).use)
 	image_info(infoing, input, FRAME(frames, i).image, colormap_infoing);
-  }
+  }*/
   
   Gif_DeleteStream(input);
   input = 0;
@@ -619,7 +661,7 @@ do_colormap_change(Gif_Stream *gfs)
 	  any_locals = 1;
       hist = histogram(gfs, &nhist);
       if (nhist <= active_output_data.colormap_size && !any_locals) {
-	warning("trivial adaptive palette (only %d colors in source)", nhist);
+	warncontext("trivial adaptive palette (only %d colors in source)", nhist);
 	return;
       }
     }
@@ -727,6 +769,47 @@ merge_and_write_frames(char *outfile, int f1, int f2)
   if (verbosing) verbose_close(']');
 }
 
+static void
+output_information(const char *outfile)
+{
+  FILE *f;
+  int i, j;
+  Gt_Frame *fr;
+  Gif_Stream *gfs;
+  
+  if (infoing == 2)
+    f = stderr;
+  else if (outfile == 0)
+    f = stdout;
+  else {
+    f = fopen(outfile, "w");
+    if (!f) {
+      error("%s: %s", outfile, strerror(errno));
+      return;
+    }
+  }
+  
+  for (i = 0; i < frames->count; i++)
+    FRAME(frames, i).stream->userflags = 97;
+
+  for (i = 0; i < frames->count; i++)
+    if (FRAME(frames, i).stream->userflags == 97) {
+      fr = &FRAME(frames, i);
+      gfs = fr->stream;
+      gfs->userflags = 0;
+      stream_info(f, gfs, fr->input_filename, fr->colormap_info,
+		  fr->extensions_info);
+      for (j = i; j < frames->count; j++)
+	if (FRAME(frames, j).stream == gfs) {
+	  fr = &FRAME(frames, j);
+	  image_info(f, gfs, fr->image, fr->colormap_info);
+	}
+    }
+  
+  if (f != stderr && f != stdout)
+    fclose(f);
+}
+
 void
 output_frames(void)
 {
@@ -738,7 +821,11 @@ output_frames(void)
   char *outfile = active_output_data.output_name;
   active_output_data.output_name = 0;
   
-  if (outputing && frames->count > 0)
+  /* Output information only now. */
+  if (infoing)
+    output_information(outfile);
+  
+  if (infoing != 1 && frames->count > 0)
     switch (mode) {
       
      case MERGING:
@@ -746,27 +833,36 @@ output_frames(void)
       merge_and_write_frames(outfile, 0, -1);
       break;
       
-     case EXPLODING:
-      /* Use the current output name for consistency, even though that means we
-	 can't explode different frames to different names. Not a big deal
-	 anyway; they can always repeat the gif on the cmd line. */
-      if (!outfile) /* Watch out! */
-	outfile = "-";
-      
-      for (i = 0; i < frames->count; i++) {
-	Gt_Frame *fr = &FRAME(frames, i);
-	int imagenumber = Gif_ImageNumber(fr->stream, fr->image);
-	char *explodename;
-	
-	char *imagename = 0;
-	if (fr->explode_by_name)
-	  imagename = fr->name ? fr->name : fr->image->identifier;
-	
-	explodename = explode_filename(outfile, imagenumber, imagename);
-	merge_and_write_frames(explodename, i, i);
-      }
-      break;
-      
+     case EXPLODING: {
+       /* Use the current output name for consistency, even though that means
+	  we can't explode different frames to different names. Not a big deal
+	  anyway; they can always repeat the gif on the cmd line. */
+       int max_nimages = 0;
+       for (i = 0; i < frames->count; i++) {
+	 Gt_Frame *fr = &FRAME(frames, i);
+	 if (fr->stream->nimages > max_nimages)
+	   max_nimages = fr->stream->nimages;
+       }
+       
+       if (!outfile) /* Watch out! */
+	 outfile = "-";
+       
+       for (i = 0; i < frames->count; i++) {
+	 Gt_Frame *fr = &FRAME(frames, i);
+	 int imagenumber = Gif_ImageNumber(fr->stream, fr->image);
+	 char *explodename;
+	 
+	 char *imagename = 0;
+	 if (fr->explode_by_name)
+	   imagename = fr->name ? fr->name : fr->image->identifier;
+	 
+	 explodename = explode_filename(outfile, imagenumber, imagename,
+					max_nimages);
+	 merge_and_write_frames(explodename, i, i);
+       }
+       break;
+     }
+     
      case INSERTING:
       /* do nothing */
       break;
@@ -787,15 +883,21 @@ output_frames(void)
  * parsing arguments
  **/
 
-void
+int
 frame_argument(Clp_Parser *clp, char *arg)
 {
-  if (parse_frame_spec(clp, arg, 1, 0) > 0) {
+  /* Returns 0 iff you should try a file named `arg'. */
+  int val = parse_frame_spec(clp, arg, -1, 0);
+  if (val == -97)
+    return 0;
+  else if (val > 0) {
     int i;
     for (i = frame_spec_1; i <= frame_spec_2; i++)
       show_frame(i, frame_spec_name != 0);
     if (next_output) combine_output_options();
-  }
+    return 1;
+  } else
+    return 1;
 }
 
 static int
@@ -808,8 +910,8 @@ handle_extension(Clp_Parser *clp, int is_app)
     Clp_OptionError(clp, "%O requires two arguments");
     return 0;
   }
-  
-  MARK_CH(frame, CH_EXTENSION);
+
+  UNCHECKED_MARK_CH(frame, CH_EXTENSION);
   if (is_app)
     gfex = Gif_NewExtension(255, extension_type);
   else if (!isdigit(extension_type[0]) && extension_type[1] == 0)
@@ -851,6 +953,7 @@ initialize_def_frame(void)
   def_frame.transparent.haspixel = 0;
   def_frame.left = -1;
   def_frame.top = -1;
+  def_frame.position_is_offset = 0;
   
   def_frame.crop = 0;
   
@@ -940,8 +1043,9 @@ redundant_option_warning(const char *option_type)
   static int context = 0;
   warning("redundant %s option", option_type);
   if (!context) {
-    warning("(An option is redundant if it had no effect before");
-    warning("being overridden by another option of the same type.)");
+    warncontext("(The %s option was overridden by another %s option",
+		option_type, option_type);
+    warncontext("before it had any effect.)");
   }
   context = 1;
 }
@@ -956,7 +1060,7 @@ print_useless_options(const char *type_name, int value, const char *names[])
     if (CHANGED(value, i)) {
       warning("useless %s-related %s option", names[i], type_name);
       if (!explanation_printed)
-	warning("(It didn't affect any %s.)", type_name);
+	warncontext("(It didn't affect any %s.)", type_name);
       explanation_printed = 1;
     }
 }
@@ -1047,36 +1151,29 @@ main(int argc, char **argv)
       /* INFORMATION OPTIONS */
       
      case INFO_OPT:
-      if (clp->negated) {
-	outputing = 1;
+      if (clp->negated)
 	infoing = 0;
-      } else {
-	outputing = !outputing;
-	infoing = outputing ? stderr : stdout;
-      }
+      else
+	/* switch between infoing == 1 (suppress regular output) and 2 (don't
+           suppress) */
+	infoing = (infoing == 1 ? 2 : 1);
       break;
       
      case COLOR_INFO_OPT:
       if (clp->negated)
-	colormap_infoing = 0;
+	def_frame.colormap_info = 0;
       else {
-	colormap_infoing = 1;
-	if (!infoing) {
-	  outputing = 0;
-	  infoing = stdout;
-	}
+	def_frame.colormap_info = 1;
+	if (!infoing) infoing = 1;
       }
       break;
       
      case EXTENSION_INFO_OPT:
       if (clp->negated)
-	extension_infoing = 0;
+	def_frame.extensions_info = 0;
       else {
-	extension_infoing = 1;
-	if (!infoing) {
-	  outputing = 0;
-	  infoing = stdout;
-	}
+	def_frame.extensions_info = 1;
+	if (!infoing) infoing = 1;
       }
       break;
       
@@ -1204,7 +1301,6 @@ main(int argc, char **argv)
      case CROP_OPT:
       if (clp->negated) goto no_crop;
       MARK_CH(frame, CH_CROP);
-      next_frame |= CH_CROP;
       {
 	Gt_Crop *crop = Gif_New(Gt_Crop);
 	/* Memory leak on crops, but this just is NOT a problem. */
@@ -1226,7 +1322,6 @@ main(int argc, char **argv)
       /* extensions options */
       
      case NO_EXTENSIONS_OPT:
-      MARK_CH(frame, CH_EXTENSION);
       def_frame.no_extensions = 1;
       break;
       
@@ -1396,8 +1491,8 @@ main(int argc, char **argv)
       MARK_CH(output, CH_RESIZE);
       if (clp->negated)
 	def_output_data.scaling = 0;
-      else if (dimensions_x <= 0 || dimensions_y <= 0) {
-	error("`--resize' width and height must be positive");
+      else if (dimensions_x <= 0 && dimensions_y <= 0) {
+	error("one of `--resize' width or height must be positive");
 	def_output_data.scaling = 0;
       } else {
 	def_output_data.scaling = 1; /* use resize dimensions */
@@ -1422,9 +1517,21 @@ main(int argc, char **argv)
       
       /* RANDOM OPTIONS */
       
+     case NO_WARNINGS_OPT:
+      no_warnings = !clp->negated;
+      break;
+      
+     case WARNINGS_OPT:
+      no_warnings = clp->negated;
+      break;
+      
      case VERSION_OPT:
+#ifdef GIF_UNGIF
+      printf("LCDF Gifsicle %s (ungif)\n", VERSION);
+#else
       printf("LCDF Gifsicle %s\n", VERSION);
-      printf("Copyright (C) 1997-9 Eddie Kohler\n\
+#endif
+      printf("Copyright (C) 1997-2000 Eddie Kohler\n\
 This is free software; see the source for copying conditions.\n\
 There is NO warranty, not even for merchantability or fitness for a\n\
 particular purpose.\n");
@@ -1447,9 +1554,7 @@ particular purpose.\n");
       /* NONOPTIONS */
       
      case Clp_NotOption:
-      if (clp->arg[0] == '#')
-	frame_argument(clp, clp->arg);
-      else {
+      if (clp->arg[0] != '#' || !frame_argument(clp, clp->arg)) {
 	input_done();
 	input_stream(clp->arg);
       }
