@@ -48,8 +48,13 @@ typedef struct Gt_Frame {
   
   unsigned flip_horizontal: 1;
   unsigned flip_vertical: 1;
+  unsigned colormap_info: 1;
+  unsigned extensions_info: 1;
+  unsigned position_is_offset: 1;
   unsigned rotation;
   
+  const char *input_filename;
+  
 } Gt_Frame;
 
 
@@ -118,9 +123,11 @@ typedef struct {
 extern const char *program_name;
 extern int verbosing;
 extern int error_count;
+extern int no_warnings;
 
 void fatal_error(char *message, ...) NORETURN;
 void warning(char *message, ...);
+void warncontext(char *message, ...);
 void error(char *message, ...);
 void clp_error_handler(char *clp_message);
 void usage(void);
@@ -137,10 +144,11 @@ void verbose_endline(void);
 /*****
  * info &c
  **/
-void stream_info(FILE *, Gif_Stream *, char *, int colormaps, int extensions);
+void stream_info(FILE *, Gif_Stream *, const char *,
+		 int colormaps, int extensions);
 void image_info(FILE *, Gif_Stream *, Gif_Image *, int colormaps);
 
-char *explode_filename(char *filename, int number, char *name);
+char *explode_filename(char *filename, int number, char *name, int max_nimg);
 
 /*****
  * merging images
