@@ -1,5 +1,5 @@
 /* gifview.c - gifview's main loop.
-   Copyright (C) 1997-9 Eddie Kohler, eddietwo@lcs.mit.edu
+   Copyright (C) 1997-2000 Eddie Kohler, eddietwo@lcs.mit.edu
    This file is part of gifview, in the gifsicle package.
 
    Gifview is free software. It is distributed under the GNU Public License,
@@ -96,31 +96,36 @@ typedef struct Gt_Viewer {
   int depth;
   Colormap colormap;
   Gif_XContext *gfx;
-
+  
   Window parent;
   int top_level;
   
   Window window;
+  int use_window;
   int width;
   int height;
-  Pixmap pixmap;
+  int resizable;
   int being_deleted;
   
   Gif_Stream *gfs;
   char *name;
   
-  Gif_Stream *anim_gfs;
-  int can_animate;
-  
   Gif_Image **im;
   int *im_number;
   int nim;
   int im_cap;
+  
+  Pixmap pixmap;
   int im_pos;
+  int was_unoptimized;
+  
+  Pixmap *unoptimized_pixmaps;
   
   struct Gt_Viewer *next;
   
+  int can_animate;
   int animating;
+  int unoptimizing;
   int scheduled;
   struct Gt_Viewer *anim_next;
   struct timeval timer;
@@ -135,6 +140,7 @@ static Display *cur_display = 0;
 static char *cur_geometry_spec = 0;
 static const char *cur_resource_name;
 static Window cur_use_window = None;
+static const char *cur_background_color = "black";
 
 static Gt_Viewer *viewers;
 static Gt_Viewer *animations;
@@ -154,9 +160,14 @@ static int interactive = 1;
 #define WINDOW_OPT		307
 #define INSTALL_COLORMAP_OPT	308
 #define INTERACTIVE_OPT		309
+#define BACKGROUND_OPT		310
+
+#define WINDOW_TYPE		(Clp_MaxDefaultType + 1)
 
 Clp_Option options[] = {
   { "animate", 'a', ANIMATE_OPT, 0, Clp_Negate },
+  { "background", 'b', BACKGROUND_OPT, Clp_ArgString, 0 },
+  { "bg", 0, BACKGROUND_OPT, Clp_ArgString, 0 },
   { "display", 'd', DISPLAY_OPT, Clp_ArgStringNotOption, 0 },
   { "geometry", 'g', GEOMETRY_OPT, Clp_ArgString, 0 },
   { "install-colormap", 'i', INSTALL_COLORMAP_OPT, 0, Clp_Negate },
@@ -165,7 +176,7 @@ Clp_Option options[] = {
   { "name", 0, NAME_OPT, Clp_ArgString, 0 },
   { "unoptimize", 'U', UNOPTIMIZE_OPT, 0, Clp_Negate },
   { "version", 0, VERSION_OPT, 0, 0 },
-  { "window", 'w', WINDOW_OPT, Clp_ArgInt, 0 },
+  { "window", 'w', WINDOW_OPT, WINDOW_TYPE, 0 },
 };
 
 
@@ -216,7 +227,7 @@ void
 usage(void)
 {
   printf("\
-`Gifview' is a lightweight GIF viewer for X. It can display animaged GIFs as\n\
+`Gifview' is a lightweight GIF viewer for X. It can display animated GIFs as\n\
 slideshows, one frame at a time, or as animations.\n\
 \n\
 Usage: %s [--display DISPLAY] [OPTION]... [FILE | FRAME]...\n\
@@ -229,7 +240,8 @@ Options are:\n\
   -g, --geometry GEOMETRY       Set window geometry.\n\
   -w, --window WINDOW           Show GIF in existing WINDOW.\n\
   -i, --install-colormap        Use a private colormap.\n\
-  +e, --no-interactive          Don't pay attention to buttons or keys.\n\
+  --bg, --background COLOR      Use COLOR for transparent pixels.\n\
+  +e, --no-interactive          Ignore buttons and keystrokes.\n\
       --help                    Print this message and exit.\n\
       --version                 Print version number and exit.\n\
 \n\
@@ -239,7 +251,8 @@ Keystrokes:\n\
   [Space] Go to next frame.             [B] Go to previous frame.\n\
   [R]/[<] Go to first frame.            [>] Go to last frame.\n\
   [ESC] Stop animation.                 [S]/[A] Toggle animation.\n\
-  [Backspace]/[W] Delete window.        [Q] Quit.\n\
+  [U] Toggle unoptimization.            [Backspace]/[W] Delete window.\n\
+  [Q] Quit.\n\
 \n\
 Left mouse button goes to next frame, right mouse button deletes window.\n\
 \n\
@@ -331,6 +344,7 @@ Gt_Viewer *
 new_viewer(Display *display, Window use_window, Gif_Stream *gfs, char *name)
 {
   Gt_Viewer *viewer;
+  int i;
   
   /* Make the Gt_Viewer structure */
   viewer = Gif_New(Gt_Viewer);
@@ -338,7 +352,12 @@ new_viewer(Display *display, Window use_window, Gif_Stream *gfs, char *name)
   
   if (use_window) {
     XWindowAttributes attr;
-    int i;
+
+    if (use_window == -1) {	/* means use root window */
+      viewer->screen_number = DefaultScreen(display);
+      use_window = RootWindow(display, viewer->screen_number);
+    }
+    
     XGetWindowAttributes(display, use_window, &attr);
     
     viewer->screen_number = -1;
@@ -355,31 +374,66 @@ new_viewer(Display *display, Window use_window, Gif_Stream *gfs, char *name)
       (display, viewer->screen_number, viewer->visual, viewer->depth,
        viewer->colormap);
     viewer->gfx->refcount++;
-    
+
+    /* use root window, if that's what we were given; otherwise, create a
+       child of the window we were given */
+    viewer->window = (use_window == attr.root ? use_window : None);
     viewer->parent = use_window;
+    viewer->use_window = (use_window == attr.root ? 1 : 0);
     viewer->top_level = 0;
+    viewer->resizable = 0;
     
   } else {
     viewer->screen_number = DefaultScreen(display);
     choose_visual(viewer);
+    viewer->window = None;
     viewer->parent = RootWindow(display, viewer->screen_number);
+    viewer->use_window = 0;
     viewer->top_level = 1;
+    viewer->resizable = 1;
+  }
+  
+  /* assign background color */
+  if (cur_background_color) {
+    XColor color;
+    if (!XParseColor(viewer->display, viewer->colormap, cur_background_color,
+		     &color)) {
+      error("invalid background color `%s'", cur_background_color);
+      cur_background_color = 0;
+    } else if (!XAllocColor(viewer->display, viewer->colormap, &color))
+      warning("can't allocate background color");
+    else {
+      unsigned long pixel = color.pixel;
+      Gif_XContext *gfx = viewer->gfx;
+      if (pixel != gfx->transparent_pixel && gfx->refcount > 1) {
+	/* copy X context */
+	viewer->gfx = Gif_NewXContextFromVisual
+	  (gfx->display, gfx->screen_number, gfx->visual, gfx->depth,
+	   gfx->colormap);
+	viewer->gfx->refcount++;
+	gfx->refcount--;
+      }
+      viewer->gfx->transparent_pixel = pixel;
+    }
   }
   
-  viewer->window = None;
   viewer->being_deleted = 0;
-  viewer->pixmap = None;
   viewer->gfs = gfs;
   viewer->name = name;
-  viewer->anim_gfs = 0;
   viewer->im_cap = Gif_ImageCount(gfs);
   viewer->im = Gif_NewArray(Gif_Image *, viewer->im_cap);
   viewer->im_number = Gif_NewArray(int, viewer->im_cap);
   viewer->nim = 0;
+  viewer->pixmap = None;
   viewer->im_pos = -1;
+  viewer->was_unoptimized = 0;
+  viewer->unoptimized_pixmaps = Gif_NewArray(Pixmap, viewer->im_cap);
+  for (i = 0; i < viewer->im_cap; i++)
+    viewer->unoptimized_pixmaps[i] = None;
   viewer->next = viewers;
   viewers = viewer;
   viewer->animating = 0;
+  viewer->unoptimizing = unoptimizing;
   viewer->scheduled = 0;
   viewer->anim_next = 0;
   viewer->anim_loop = 0;
@@ -392,7 +446,12 @@ void
 delete_viewer(Gt_Viewer *viewer)
 {
   Gt_Viewer *prev = 0, *trav;
-  if (viewer->pixmap) XFreePixmap(viewer->display, viewer->pixmap);
+  int i;
+  if (viewer->pixmap && !viewer->was_unoptimized)
+    XFreePixmap(viewer->display, viewer->pixmap);
+  for (i = 0; i < viewer->im_cap; i++)
+    if (viewer->unoptimized_pixmaps[i])
+      XFreePixmap(viewer->display, viewer->unoptimized_pixmaps[i]);
   
   for (trav = viewers; trav != viewer; prev = trav, trav = trav->next)
     ;
@@ -400,8 +459,6 @@ delete_viewer(Gt_Viewer *viewer)
   else viewers = viewer->next;
   
   Gif_DeleteStream(viewer->gfs);
-  if (viewer->anim_gfs != viewer->gfs)
-    Gif_DeleteStream(viewer->anim_gfs);
   Gif_DeleteArray(viewer->im);
   Gif_DeleteArray(viewer->im_number);
   Gif_DeleteXContext(viewer->gfx);
@@ -426,7 +483,7 @@ get_input_stream(char *name)
     return 0;
   }
   
-  gfs = Gif_FullReadFile(f, GIF_READ_COMPRESSED);
+  gfs = Gif_FullReadFile(f, GIF_READ_COMPRESSED, 0, 0);
   fclose(f);
   
   if (!gfs || Gif_ImageCount(gfs) == 0) {
@@ -443,13 +500,9 @@ get_input_stream(char *name)
     }
   }
   
-  if (unoptimizing)
-    Gif_Unoptimize(gfs);
-  
   viewer = new_viewer(cur_display, cur_use_window, gfs, name);
-  cur_use_window = None;
-  if (unoptimizing)
-    viewer->anim_gfs = gfs;
+  if (cur_use_window)
+    cur_use_window = None;
   return viewer;
 }
 
@@ -462,19 +515,11 @@ void
 switch_animating(Gt_Viewer *viewer, int animating)
 {
   int i;
-  Gif_Stream *gfs;
-  
+  Gif_Stream *gfs = viewer->gfs;
   if (animating == viewer->animating || !viewer->can_animate)
     return;
-  
-  if (animating && !viewer->anim_gfs) {
-    viewer->anim_gfs = Gif_CopyStreamImages(viewer->gfs);
-    Gif_Unoptimize(viewer->anim_gfs);
-  }
-  gfs = animating ? viewer->anim_gfs : viewer->gfs;
   for (i = 0; i < gfs->nimages; i++)
     viewer->im[i] = gfs->images[i];
-  
   viewer->animating = animating;
 }
 
@@ -584,7 +629,6 @@ void
 create_viewer_window(Gt_Viewer *viewer, int w, int h)
 {
   Display *display = viewer->display;
-  Window window;
   char *stringlist[2];
   XTextProperty window_name, icon_name;
   XClassHint classh;
@@ -600,7 +644,7 @@ create_viewer_window(Gt_Viewer *viewer, int w, int h)
   }
   
   /* Open the display and create the window */
-  {
+  if (!viewer->window) {
     XSetWindowAttributes x_set_attr;
     unsigned long x_set_attr_mask;
     x_set_attr.colormap = viewer->colormap;
@@ -611,21 +655,18 @@ create_viewer_window(Gt_Viewer *viewer, int w, int h)
     x_set_attr_mask = CWColormap | CWBorderPixel | CWBackPixel
       | CWBackingStore | CWSaveUnder;
     
-    viewer->window = window = XCreateWindow
+    viewer->window = XCreateWindow
       (display, viewer->parent,
        sizeh->x, sizeh->y, sizeh->width, sizeh->height, 0,
        viewer->depth, InputOutput, viewer->visual,
        x_set_attr_mask, &x_set_attr);
   }
-  
+
+  /* If user gave us geometry, don't change the size later */
   if (sizeh->flags & USSize)
-    /* Setting sizes to -1 says user gave us geometry, don't change the size
-       later */
-    viewer->width = viewer->height = -1;
-  else {
-    viewer->width = w;
-    viewer->height = h;
-  }
+    viewer->resizable = 0;
+  viewer->width = w;
+  viewer->height = h;
   
   /* Set the window's title and class (for window manager resources) */
   if (viewer->top_level) {
@@ -635,7 +676,7 @@ create_viewer_window(Gt_Viewer *viewer, int w, int h)
     XStringListToTextProperty(stringlist, 1, &icon_name);
     classh.res_name = (char *)cur_resource_name;
     classh.res_class = "Gifview";
-    XSetWMProperties(display, window, &window_name, &icon_name,
+    XSetWMProperties(display, viewer->window, &window_name, &icon_name,
 		     NULL, 0, sizeh, NULL, &classh);
     XFree(window_name.value);
     XFree(icon_name.value);
@@ -644,14 +685,14 @@ create_viewer_window(Gt_Viewer *viewer, int w, int h)
       wm_delete_window_atom = XInternAtom(display, "WM_DELETE_WINDOW", False);
       wm_protocols_atom = XInternAtom(display, "WM_PROTOCOLS", False);
     }
-    XSetWMProtocols(display, window, &wm_delete_window_atom, 1);
+    XSetWMProtocols(display, viewer->window, &wm_delete_window_atom, 1);
   }
 
   if (interactive)
-    XSelectInput(display, window, ButtonPressMask | KeyPressMask
+    XSelectInput(display, viewer->window, ButtonPressMask | KeyPressMask
 		 | StructureNotifyMask);
   else
-    XSelectInput(display, window, StructureNotifyMask);
+    XSelectInput(display, viewer->window, StructureNotifyMask);
   
   XFree(sizeh);
 }
@@ -665,7 +706,7 @@ pre_delete_viewer(Gt_Viewer *viewer)
   
   if (viewer->scheduled) unschedule(viewer);
   
-  if (viewer->window)
+  if (viewer->window && !viewer->use_window)
     XDestroyWindow(viewer->display, viewer->window);
   else
     delete_viewer(viewer);
@@ -721,6 +762,27 @@ set_viewer_name(Gt_Viewer *viewer)
 }
 
 
+static Pixmap
+unoptimized_frame(Gt_Viewer *viewer, int frame)
+{
+  /* Never frees viewer->pixmap. */
+  int i;
+  
+  if (!viewer->unoptimized_pixmaps[frame]) {
+    for (i = 0; i < frame; i++)
+      if (!viewer->unoptimized_pixmaps[i])
+	unoptimized_frame(viewer, i);
+    
+    viewer->unoptimized_pixmaps[frame] =
+      Gif_XNextImage(viewer->gfx,
+		     (frame < 2 ? None : viewer->unoptimized_pixmaps[frame-2]),
+		     (frame < 1 ? None : viewer->unoptimized_pixmaps[frame-1]),
+		     viewer->gfs, frame);
+  }
+
+  return viewer->unoptimized_pixmaps[frame];
+}
+
 void
 view_frame(Gt_Viewer *viewer, int frame)
 {
@@ -749,37 +811,30 @@ view_frame(Gt_Viewer *viewer, int frame)
   
   if (frame != viewer->im_pos) {
     Gif_Image *gfi = viewer->im[frame];
-    int width = Gif_ImageWidth(gfi);
-    int height = Gif_ImageHeight(gfi);
+    int width, height;
     
-    viewer->im_pos = frame;
+    /* 5/26/98 Do some noodling around to try and use memory most effectively.
+       If animating, keep the uncompressed frame; otherwise, throw it away. */
+    if (viewer->animating || viewer->unoptimizing)
+      viewer->pixmap = unoptimized_frame(viewer, frame);
+    else
+      viewer->pixmap = Gif_XImage(viewer->gfx, viewer->gfs, gfi);
     
+    /* put the image on the window */
+    if (viewer->animating || viewer->unoptimizing)
+      width = viewer->gfs->screen_width, height = viewer->gfs->screen_height;
+    else
+      width = Gif_ImageWidth(gfi), height = Gif_ImageHeight(gfi);
     if (!window) {
       create_viewer_window(viewer, width, height);
       window = viewer->window;
     }
-    
-    if ((!viewer->animating && Gif_ImageCount(viewer->gfs) > 1)
-	|| old_pixmap == None)
-      need_set_name = 1;
-    
-    /* 5/26/98 Do some noodling around to try and use memory most effectively.
-       If animating, keep the uncompressed frame; otherwise, throw it away. */
-    Gif_UncompressImage(gfi);
-    viewer->pixmap = Gif_XImage(viewer->gfx, viewer->gfs, gfi);
-    if (!viewer->animating && gfi->compressed)
-      Gif_ReleaseUncompressedImage(gfi);
-    
     XSetWindowBackgroundPixmap(display, window, viewer->pixmap);
-    if (old_pixmap) {
+    if (old_pixmap || viewer->use_window) /* clear if using existing window */
       XClearWindow(display, window);
-      XFreePixmap(display, old_pixmap);
-    }
-
     /* Only change size after changing pixmap. */
     if ((viewer->width != width || viewer->height != height)
-	&& viewer->width > 0) {
-      /* If viewer->width < 0, user gave geometry; don't change window size */
+	&& viewer->resizable) {
       XWindowChanges winch;
       winch.width = viewer->width = width;
       winch.height = viewer->height = height;
@@ -787,12 +842,24 @@ view_frame(Gt_Viewer *viewer, int frame)
 	(display, window, viewer->screen_number,
 	 CWWidth | CWHeight, &winch);
     }
+    
+    /* Get rid of old pixmaps */
+    if (old_pixmap && !viewer->was_unoptimized)
+      XFreePixmap(display, old_pixmap);
+    viewer->was_unoptimized = viewer->animating || viewer->unoptimizing;
+    
+    /* Do we need a new name? */
+    if ((!viewer->animating && Gif_ImageCount(viewer->gfs) > 1)
+	|| old_pixmap == None)
+      need_set_name = 1;
+    
+    viewer->im_pos = frame;
   }
   
   if (need_set_name)
     set_viewer_name(viewer);
   
-  if (!old_pixmap)
+  if (!old_pixmap && !viewer->use_window)
     /* first image; map the window */
     XMapRaised(display, window);
   else if (viewer->animating)
@@ -926,6 +993,16 @@ key_press(Gt_Viewer *viewer, XKeyEvent *e)
     
     set_viewer_name(viewer);
     
+  } else if (key == XK_U || key == XK_u) {
+    /* U: toggle unoptimizing */
+    int pos = viewer->im_pos;
+    viewer->unoptimizing = !viewer->unoptimizing;
+    if (!viewer->animating) {
+      viewer->im_pos = -1;
+      view_frame(viewer, pos);
+      set_viewer_name(viewer);
+    }
+    
   } else if (key == XK_R || key == XK_r
 	     || (nbuf == 1 && buf[0] == '<')) {
     /* R or <: reset to first frame */
@@ -944,6 +1021,10 @@ key_press(Gt_Viewer *viewer, XKeyEvent *e)
     switch_animating(viewer, 0);
     unschedule(viewer);
     set_viewer_name(viewer);
+    
+  } else if (key == XK_Z || key == XK_z) {
+    /* Z: trigger resizability */
+    viewer->resizable = !viewer->resizable;
   }
 }
 
@@ -1037,6 +1118,10 @@ main(int argc, char **argv)
     Clp_NewParser(argc, argv,
                   sizeof(options) / sizeof(options[0]), options);
   Clp_SetOptionChar(clp, '+', Clp_ShortNegated);
+  Clp_AddStringListType
+    (clp, WINDOW_TYPE, Clp_AllowNumbers,
+     "root", -1,
+     0);
   program_name = cur_resource_name = Clp_ProgramName(clp);
   
   xwGETTIMEOFDAY(&genesis_time);
@@ -1064,10 +1149,14 @@ main(int argc, char **argv)
       unoptimizing = clp->negated ? 0 : 1;
       break;
       
+     case BACKGROUND_OPT:
+      cur_background_color = clp->arg;
+      break;
+      
      case ANIMATE_OPT:
       animating = clp->negated ? 0 : 1;
       break;
-
+      
      case INSTALL_COLORMAP_OPT:
       install_colormap = clp->negated ? 0 : 1;
       break;
@@ -1075,14 +1164,14 @@ main(int argc, char **argv)
      case WINDOW_OPT:
       cur_use_window = clp->val.u;
       break;
-
+      
      case INTERACTIVE_OPT:
       interactive = clp->negated ? 0 : 1;
       break;
       
      case VERSION_OPT:
       printf("gifview (LCDF Gifsicle) %s\n", VERSION);
-      printf("Copyright (C) 1997-9 Eddie Kohler\n\
+      printf("Copyright (C) 1997-2000 Eddie Kohler\n\
 This is free software; see the source for copying conditions.\n\
 There is NO warranty, not even for merchantability or fitness for a\n\
 particular purpose.\n");
