@@ -17,11 +17,14 @@ typedef struct {
   u_int16_t top;
   u_int16_t width;
   u_int16_t height;
+  u_int32_t size;
   byte disposal;
   int transparent;
   byte *needed_colors;
   u_int16_t required_color_count;
-  u_int16_t global_penalty;
+  int32_t active_penalty;
+  int32_t global_penalty;
+  int32_t colormap_penalty;
   Gif_Image *new_gfi;
 } Gif_OptData;
 
@@ -223,8 +226,9 @@ apply_frame(u_int16_t *dst, Gif_Image *gfi, int replace)
   /* make sure transparency maps to TRANSP */
   for (i = 0; i < colormap->ncol; i++)
     map[i] = colormap->col[i].pixel;
+  /* out-of-bounds colors map to 0, for the sake of argument */
   for (i = colormap->ncol; i < 256; i++)
-    map[i] = TRANSP;
+    map[i] = colormap->col[0].pixel;
   if (gfi->transparent >= 0 && gfi->transparent < 256)
     map[gfi->transparent] = TRANSP;
   else
@@ -270,15 +274,30 @@ apply_frame_disposal(u_int16_t *into_data, u_int16_t *from_data,
    the changes and store it in `bounds'. */
 
 static void
-find_difference_bounds(Gif_OptData *bounds)
+find_difference_bounds(Gif_OptData *bounds, Gif_Image *gfi, Gif_Image *last)
 {
-  int lf, rt, tp, bt, x, y;
+  int lf, rt, lf_min, rt_max, tp, bt, x, y;
   
-  for (tp = 0; tp < screen_height; tp++)
+  /* 1.Aug.99 - use current bounds if possible, since this function is a speed
+     bottleneck */
+  if (!last || last->disposal == GIF_DISPOSAL_NONE
+      || last->disposal == GIF_DISPOSAL_ASIS) {
+    lf_min = gfi->left;
+    rt_max = gfi->left + gfi->width - 1;
+    tp = gfi->top;
+    bt = gfi->top + gfi->height - 1;
+  } else {
+    lf_min = 0;
+    rt_max = screen_width - 1;
+    tp = 0;
+    bt = screen_height - 1;
+  }
+  
+  for (; tp < screen_height; tp++)
     if (memcmp(last_data + screen_width * tp, this_data + screen_width * tp,
 	       screen_width * sizeof(u_int16_t)) != 0)
       break;
-  for (bt = screen_height - 1; bt >= tp; bt--)
+  for (; bt >= tp; bt--)
     if (memcmp(last_data + screen_width * bt, this_data + screen_width * bt,
 	       screen_width * sizeof(u_int16_t)) != 0)
       break;
@@ -288,16 +307,20 @@ find_difference_bounds(Gif_OptData *bounds)
   for (y = tp; y <= bt; y++) {
     u_int16_t *ld = last_data + screen_width * y;
     u_int16_t *td = this_data + screen_width * y;
-    for (x = 0; x < lf; x++)
+    for (x = lf_min; x < lf; x++)
       if (ld[x] != td[x])
 	break;
     lf = x;
     
-    for (x = screen_width - 1; x > rt; x--)
+    for (x = rt_max; x > rt; x--)
       if (ld[x] != td[x])
 	break;
     rt = x;
   }
+
+  /* 19.Aug.1999 - handle case when there's no difference between frames */
+  if (tp > bt)
+    tp = bt = lf = rt = 0;
   
   bounds->left = lf;
   bounds->top = tp;
@@ -530,7 +553,7 @@ create_subimages(Gif_Stream *gfs, int optimize_level)
     /* find minimum area of difference between this image and last image */
     subimage->disposal = GIF_DISPOSAL_ASIS;
     if (image_index > 0)
-      find_difference_bounds(subimage);
+      find_difference_bounds(subimage, gfi, last_gfi);
     else {
       subimage->left = gfi->left;
       subimage->top = gfi->top;
@@ -564,7 +587,7 @@ create_subimages(Gif_Stream *gfs, int optimize_level)
       get_used_colors(subimage, use_transparency);
     }
     
-    gfi->userdata = subimage;
+    gfi->user_data = subimage;
     last_gfi = gfi;
     
     /* Apply optimized disposal to last_data and unoptimized disposal to
@@ -594,205 +617,129 @@ create_subimages(Gif_Stream *gfs, int optimize_level)
  * CALCULATE OUTPUT GLOBAL COLORMAP
  **/
 
-/* choose_256_colors: If we need to have some local colormaps, then choose for
-   the global colormap an optimal subset of all the colors, to minimize the
-   size of extra local colormaps required.
-   
-   On return, the `global_penalty' component of an image's Gif_OptData
-   structure is 0 iff that image will need a local colormap. */
-
-static void
-choose_256_colors(Gif_Stream *gfs, u_int16_t *global_all)
-{
-  int i, imagei;
-  int all_ncol = all_colormap->ncol;
-  int32_t *penalty = Gif_NewArray(int32_t, all_ncol);
-  u_int16_t *ordering = Gif_NewArray(u_int16_t, all_ncol);
-  int nordering = all_ncol - 1;
-  int penalties_changed;
-  
-  for (i = 0; i < nordering; i++)
-    ordering[i] = i + 1;
-  
-  /* choose appropriate penalties for each image */
-  for (imagei = 0; imagei < gfs->nimages; imagei++) {
-    Gif_OptData *optdata = (Gif_OptData *)gfs->images[imagei]->userdata;
-    optdata->global_penalty = 1;
-    for (i = 1; i < optdata->required_color_count; i *= 2)
-      optdata->global_penalty *= 3;
-  }
-  
-  /* set initial penalties for each color */
-  for (i = 0; i < all_ncol; i++)
-    penalty[i] = 0;
-  for (imagei = 0; imagei < gfs->nimages; imagei++) {
-    Gif_OptData *optdata = (Gif_OptData *)gfs->images[imagei]->userdata;
-    byte *need = optdata->needed_colors;
-    int this_penalty = optdata->global_penalty;
-    for (i = 0; i < all_ncol; i++)
-      if (need[i] == REQUIRED)
-	penalty[i] += this_penalty;
-  }
-  
-  /* force the background to occur in the global colormap by giving it a very
-     high penalty */
-  if (background != TRANSP)
-    penalty[background] = 0x7FFFFFFF;
-  
-  /* loop, removing the most useless color each time, until we have exactly
-     256 colors */
-  penalties_changed = 1;
-  while (nordering > 256) {
-    int removed_color;
-    
-    if (penalties_changed)
-      sort_permutation(ordering, nordering, penalty, 1);
-    
-    /* remove the color which is least expensive to remove */
-    removed_color = ordering[nordering - 1];
-    nordering--;
-    
-    /* adjust penalties. if an image now must have a local colormap, then any
-       penalty values for its other colors shouldn't be considered */
-    penalties_changed = 0;
-    for (imagei = 0; imagei < gfs->nimages; imagei++) {
-      Gif_OptData *optdata = (Gif_OptData *)gfs->images[imagei]->userdata;
-      byte *need = optdata->needed_colors;
-      int this_penalty = optdata->global_penalty;
-      
-      if (this_penalty == 0 || need[removed_color] != REQUIRED)
-	continue;
-      
-      for (i = 0; i < all_ncol; i++)
-	if (need[i] == REQUIRED)
-	  penalty[i] -= this_penalty;
-      optdata->global_penalty = 0;
-      penalties_changed = 1;
-    }
-  }
-  
-  for (i = 0; i < 256; i++)
-    global_all[i] = ordering[i];
-  
-  Gif_DeleteArray(ordering);
-  Gif_DeleteArray(penalty);
-}
-
 /* create_out_global_map: The interface function to this pass. It creates
    out_global_map and sets pixel values on all_colormap appropriately.
    Specifically:
    
    all_colormap->col[P].pixel >= 256 ==> P is not in the global colormap.
    
-   Otherwise, all_colormap->col[P].pixel == some J so that
-   GIF_COLOREQ(&all_colormap->col[P], &out_global_map->col[J]). */
+   Otherwise, all_colormap->col[P].pixel == the J so that
+   GIF_COLOREQ(&all_colormap->col[P], &out_global_map->col[J]).
+
+   On return, the `colormap_penalty' component of an image's Gif_OptData
+   structure is <0 iff that image will need a local colormap.
+
+   20.Aug.1999 - updated to new version that arranges the entire colormap, not
+   just the stuff above 256 colors. */
+
+static void
+increment_penalties(Gif_OptData *opt, int32_t *penalty, int32_t delta)
+{
+  int i;
+  int all_ncol = all_colormap->ncol;
+  byte *need = opt->needed_colors;
+  for (i = 1; i < all_ncol; i++)
+    if (need[i] == REQUIRED)
+      penalty[i] += delta;
+}
 
 static void
 create_out_global_map(Gif_Stream *gfs)
 {
-  int i, imagei;
   int all_ncol = all_colormap->ncol;
-  u_int16_t *global_all = Gif_NewArray(u_int16_t, 256);
-  int nglobal_all;
-
-  /* 1. First determine which colors should be in the global colormap. Do this
-     by calculating the penalty should a color NOT be in the global colormap;
-     then choose the lowest of these.
-     
-     Note: Don't need to do this if there are <=256 colors overall. */
+  int32_t *penalty = Gif_NewArray(int32_t, all_ncol);
+  u_int16_t *permute = Gif_NewArray(u_int16_t, all_ncol);
+  u_int16_t *ordering = Gif_NewArray(u_int16_t, all_ncol);
+  int cur_ncol, i, imagei;
+  int nglobal_all = (all_ncol <= 257 ? all_ncol - 1 : 256);
+  int permutation_changed;
+		     
+  /* initial permutation is null */
+  for (i = 0; i < all_ncol - 1; i++)
+    permute[i] = i + 1;
   
-  if (all_ncol - 1 > 256) {
-    nglobal_all = 256;
-    choose_256_colors(gfs, global_all);
-  } else {
-    nglobal_all = all_ncol - 1;
-    for (i = 0; i < nglobal_all; i++)
-      global_all[i] = i + 1;
-    /* Depend on each image's global_penalty being != 0 by default */
-  }
-
-  /* Colormap Canonicalization
-     
-     Markus F.X.J. Oberhumer <k3040e4@c210.edvz.uni-linz.ac.at> would like
-     gifsicle to `canonicalize' colormaps, meaning that you might want the
-     colormap to have a predictable arrangement after piping it through
-     gifsicle. This is a nice place to do that. Basically, sort the colormap
-     on RGB order first (so that colors with equal rank have a predictable
-     order), then sort on rank. (For rank sorting information see below.) */
-  {
-    int32_t *rgb_rank = Gif_NewArray(int32_t, all_ncol);
-    Gif_Color *all_col = all_colormap->col;
-    for (i = 0; i < all_ncol; i++)
-      rgb_rank[i] = (all_col[i].red << 16) | (all_col[i].green << 8)
-	| (all_col[i].blue);
-    
-    sort_permutation(global_all, nglobal_all, rgb_rank, 0);
-    
-    Gif_DeleteArray(rgb_rank);
+  /* choose appropriate penalties for each image */
+  for (imagei = 0; imagei < gfs->nimages; imagei++) {
+    Gif_OptData *opt = (Gif_OptData *)gfs->images[imagei]->user_data;
+    opt->global_penalty = opt->colormap_penalty = 1;
+    for (i = 2; i < opt->required_color_count; i *= 2)
+      opt->colormap_penalty *= 3;
+    opt->active_penalty =
+      (all_ncol > 257 ? opt->colormap_penalty : opt->global_penalty);
   }
   
-  /* Now, reorder global colors. Colors used in a lot of images should appear
-     first in the global colormap; then those images might be able to use a
-     smaller min_code_size, since the colors they need are clustered first.
-     The strategy used here -- just count in how many images a color is used
-     and sort on that -- isn't strictly optimal, but it does well for most
-     images. */
-  {
-    int32_t *rank = Gif_NewArray(int32_t, all_ncol);
-    for (i = 0; i < all_ncol; i++)
-      rank[i] = 0;
+  /* set initial penalties for each color */
+  for (i = 1; i < all_ncol; i++)
+    penalty[i] = 0;
+  for (imagei = 0; imagei < gfs->nimages; imagei++) {
+    Gif_OptData *opt = (Gif_OptData *)gfs->images[imagei]->user_data;
+    increment_penalties(opt, penalty, opt->active_penalty);
+  }
+  permutation_changed = 1;
+  
+  /* Loop, removing one color at a time. */
+  for (cur_ncol = all_ncol - 1; cur_ncol; cur_ncol--) {
+    u_int16_t removed;
     
+    /* sort permutation based on penalty */
+    if (permutation_changed)
+      sort_permutation(permute, cur_ncol, penalty, 1);
+    permutation_changed = 0;
+    
+    /* update reverse permutation */
+    removed = permute[cur_ncol - 1];
+    ordering[removed] = cur_ncol - 1;
+    
+    /* decrement penalties for colors that are out of the running */
     for (imagei = 0; imagei < gfs->nimages; imagei++) {
-      Gif_OptData *optdata = (Gif_OptData *)gfs->images[imagei]->userdata;
-      byte *need = optdata->needed_colors;
-      /* ignore images that will require a local colormap */
-      if (!optdata->global_penalty)
-	continue;
-      for (i = 1; i < all_ncol; i++)
-	if (need[i] == REQUIRED)
-	  rank[i]++;
+      Gif_OptData *opt = (Gif_OptData *)gfs->images[imagei]->user_data;
+      byte *need = opt->needed_colors;
+      if (opt->global_penalty > 0 && need[removed] == REQUIRED) {
+	increment_penalties(opt, penalty, -opt->active_penalty);
+	opt->global_penalty = 0;
+	opt->colormap_penalty = (cur_ncol > 256 ? -1 : 0);
+	permutation_changed = 1;
+      }
     }
     
-    sort_permutation(global_all, nglobal_all, rank, 1);
-    
-    /* get rid of unused colors; warning: we may need to include the
-       background explicitly below */
-    for (i = 0; i < nglobal_all; i++)
-      if (!rank[global_all[i]])
-	nglobal_all = i;
-    
-    Gif_DeleteArray(rank);
+    /* change colormap penalties if we're no longer working w/globalmap */
+    if (cur_ncol == 257) {
+      for (i = 0; i < all_ncol; i++)
+	penalty[i] = 0;
+      for (imagei = 0; imagei < gfs->nimages; imagei++) {
+	Gif_OptData *opt = (Gif_OptData *)gfs->images[imagei]->user_data;
+	opt->active_penalty = opt->global_penalty;
+	increment_penalties(opt, penalty, opt->global_penalty);
+      }
+      permutation_changed = 1;
+    }
   }
   
-  /* Finally, make out_global_map. */
+  /* make sure background is in the global colormap */
+  if (background != TRANSP && ordering[background] >= 256) {
+    u_int16_t other = permute[255];
+    ordering[other] = ordering[background];
+    ordering[background] = 255;
+  }
+  
+  /* assign out_global_map based on permutation */
   out_global_map = Gif_NewFullColormap(nglobal_all, 256);
   
-  for (i = 0; i < all_ncol; i++)
-    all_colormap->col[i].pixel = NOT_IN_OUT_GLOBAL;
-  for (i = 0; i < nglobal_all; i++) {
-    int all_i = global_all[i];
-    out_global_map->col[i] = all_colormap->col[all_i];
-    all_colormap->col[all_i].pixel = i;
-  }
+  for (i = 1; i < all_ncol; i++)
+    if (ordering[i] < 256) {
+      out_global_map->col[ordering[i]] = all_colormap->col[i];
+      all_colormap->col[i].pixel = ordering[i];
+    } else
+      all_colormap->col[i].pixel = NOT_IN_OUT_GLOBAL;
   
   /* set the stream's background color */
-  if (background != TRANSP) {
-    int val = all_colormap->col[background].pixel;
-    if (val == NOT_IN_OUT_GLOBAL) {
-      /* this is the case where there were less than 256 colors total, but the
-	 background wasn't used, so it was removed above. there must be room
-	 for it in the output colormap. */
-      assert(out_global_map->ncol < 256);
-      val = out_global_map->ncol;
-      out_global_map->col[val] = all_colormap->col[background];
-      all_colormap->col[background].pixel = val;
-      out_global_map->ncol++;
-    }
-    gfs->background = val;
-  }
-  
-  Gif_DeleteArray(global_all);
-  assert(out_global_map->ncol <= 256);
+  if (background != TRANSP)
+    gfs->background = ordering[background];
+
+  /* cleanup */
+  Gif_DeleteArray(penalty);
+  Gif_DeleteArray(permute);
+  Gif_DeleteArray(ordering);
 }
 
 
@@ -863,16 +810,15 @@ prepare_colormap_map(Gif_Image *gfi, Gif_Colormap *into, byte *need)
 	break;
       }
     
-    /* otherwise, add another slot to `into'; it's pure transparent. Use the
-       color we put into all_colormap->col[TRANSP] earlier. If there's no room
-       for it, then give up */
+    /* otherwise, [1.Aug.1999] use a fake slot for the purely transparent
+       color. Don't actually enter the transparent color into the colormap --
+       we might be able to output a smaller colormap! If there's no room for
+       it, give up */
     if (transparent < 0) {
-      if (is_global && all_col[TRANSP].pixel < NOT_IN_OUT_GLOBAL)
-	transparent = all_col[TRANSP].pixel;
-      else if (ncol < 256) {
+      if (ncol < 256) {
 	transparent = ncol;
-	col[ncol++] = all_col[TRANSP];
-	if (is_global) all_col[TRANSP].pixel = transparent;
+	/* 1.Aug.1999 - don't increase ncol */
+	col[ncol] = all_col[TRANSP];
       } else
 	goto error;
     }
@@ -946,10 +892,15 @@ prepare_colormap(Gif_Image *gfi, byte *need)
     
     for (i = 0; i < gfi->local->ncol; i++)
       permutation[local_col[i].pixel] = i;
+    /* 1.Aug.1999 - we might not have added space for gfi->transparent */
+    if (gfi->transparent >= gfi->local->ncol)
+      permutation[gfi->transparent] = gfi->transparent;
+    
     for (i = 0; i < all_colormap->ncol; i++)
       map[i] = permutation[map[i]];
+
     if (gfi->transparent >= 0)
-      gfi->transparent = map[TRANSP] = permutation[gfi->transparent];
+      gfi->transparent = map[TRANSP];
   }
   
   return map;
@@ -1142,7 +1093,7 @@ create_new_image_data(Gif_Stream *gfs, int optimize_level)
   
   for (image_index = 0; image_index < gfs->nimages; image_index++) {
     Gif_Image *cur_gfi = gfs->images[image_index];
-    Gif_OptData *optdata = (Gif_OptData *)cur_gfi->userdata;
+    Gif_OptData *opt = (Gif_OptData *)cur_gfi->user_data;
     
     /* save previous data if necessary */
     if (cur_gfi->disposal == GIF_DISPOSAL_PREVIOUS) {
@@ -1159,16 +1110,16 @@ create_new_image_data(Gif_Stream *gfs, int optimize_level)
     
     /* set bounds and disposal from optdata */
     Gif_ReleaseUncompressedImage(cur_gfi);
-    cur_gfi->left = optdata->left;
-    cur_gfi->top = optdata->top;
-    cur_gfi->width = optdata->width;
-    cur_gfi->height = optdata->height;
-    cur_gfi->disposal = optdata->disposal;
+    cur_gfi->left = opt->left;
+    cur_gfi->top = opt->top;
+    cur_gfi->width = opt->width;
+    cur_gfi->height = opt->height;
+    cur_gfi->disposal = opt->disposal;
     if (image_index > 0) cur_gfi->interlace = 0;
     
     /* find the new image's colormap and then make new data */
     {
-      byte *map = prepare_colormap(cur_gfi, optdata->needed_colors);
+      byte *map = prepare_colormap(cur_gfi, opt->needed_colors);
       byte *data = Gif_NewArray(byte, cur_gfi->width * cur_gfi->height);
       Gif_SetUncompressedImage(cur_gfi, data, Gif_DeleteArrayFunc, 0);
       
@@ -1181,8 +1132,8 @@ create_new_image_data(Gif_Stream *gfs, int optimize_level)
       Gif_DeleteArray(map);
     }
     
-    delete_opt_data(optdata);
-    cur_gfi->userdata = 0;
+    delete_opt_data(opt);
+    cur_gfi->user_data = 0;
     
     /* Set up last_data and this_data. last_data must contain this_data + new
        disposal. this_data must contain this_data + old disposal. */
