@@ -19,6 +19,7 @@
 const char *program_name = "gifsicle";
 static int verbose_pos = 0;
 int error_count = 0;
+int no_warnings = 0;
 
 
 static void
@@ -28,15 +29,17 @@ verror(int seriousness, char *message, va_list val)
   char buffer[BUFSIZ];
   verbose_endline();
   
-  if (seriousness > 1)
+  if (seriousness > 2)
     sprintf(pattern, "%s: fatal error: %%s\n", program_name);
-  else if (seriousness == 0)
+  else if (seriousness == 1)
     sprintf(pattern, "%s: warning: %%s\n", program_name);
   else
     sprintf(pattern, "%s: %%s\n", program_name);
   
-  if (seriousness != 0)
+  if (seriousness > 1)
     error_count++;
+  else if (no_warnings)
+    return;
   
   /* try and keep error messages together (no interleaving of error messages
      from two gifsicle processes in the same command line) by calling fprintf
@@ -56,7 +59,7 @@ fatal_error(char *message, ...)
 {
   va_list val;
   va_start(val, message);
-  verror(2, message, val);
+  verror(3, message, val);
   va_end(val);
   exit(EXIT_USER_ERR);
 }
@@ -66,12 +69,21 @@ error(char *message, ...)
 {
   va_list val;
   va_start(val, message);
-  verror(1, message, val);
+  verror(2, message, val);
   va_end(val);
 }
 
 void
 warning(char *message, ...)
+{
+  va_list val;
+  va_start(val, message);
+  verror(1, message, val);
+  va_end(val);
+}
+
+void
+warncontext(char *message, ...)
 {
   va_list val;
   va_start(val, message);
@@ -123,6 +135,7 @@ General options: Also --no-OPTION for info and verbose.\n\
   -h, --help                    Print this message and exit.\n\
       --version                 Print version number and exit.\n\
   -o, --output FILE             Write output to FILE.\n\
+  -w, --no-warnings             Don't report warnings.\n\
 \n", program_name);
   printf("\
 Frame selections:               #num, #num1-num2, #num1-, #name\n\
@@ -158,7 +171,7 @@ Animation options: Also --no-OPTION and --same-OPTION.\n\
   -d, --delay TIME              Sets frame delay to TIME (in 1/100sec).\n\
   -D, --disposal METHOD         Sets frame disposal to METHOD.\n\
   -l, --loopcount[=N]           Sets loop extension to N (default forever).\n\
-  -O, -optimize[=LEV]           Optimize output GIFs.\n\
+  -O, --optimize[=LEV]          Optimize output GIFs.\n\
   -U, --unoptimize              Unoptimize input GIFs.\n\
 \n");
   printf("\
@@ -175,6 +188,11 @@ Whole-GIF options: Also --no-OPTION.\n\
 \n\
 Report bugs to <eddietwo@lcs.mit.edu>.\n\
 Too much information? Try `%s --help | more'.\n", program_name);
+#ifdef GIF_UNGIF
+  printf("\
+This version of gifsicle writes uncompressed GIFs, which can be far larger\n\
+than compressed GIFs. See http://www.lcdf.org/gifsicle for more information.\n");
+#endif
 }
 
 
@@ -331,18 +349,17 @@ extension_info(FILE *where, Gif_Stream *gfs, Gif_Extension *gfex, int count)
 
 
 void
-stream_info(FILE *where, Gif_Stream *gfs, char *filename,
+stream_info(FILE *where, Gif_Stream *gfs, const char *filename,
 	    int colormaps, int extensions)
 {
   Gif_Extension *gfex;
   int n;
   
   if (!gfs) return;
-  gfs->userflags = 0; /* clear userflags to indicate stream info produced */
   
   verbose_endline();
-  fprintf(where, "* %s %d image%s\n", filename, gfs->nimages,
-	  gfs->nimages == 1 ? "" : "s");
+  fprintf(where, "* %s %d image%s\n", (filename ? filename : "<stdin>"),
+	  gfs->nimages, gfs->nimages == 1 ? "" : "s");
   fprintf(where, "  logical screen %dx%d\n",
 	  gfs->screen_width, gfs->screen_height);
   
@@ -394,10 +411,9 @@ image_info(FILE *where, Gif_Stream *gfs, Gif_Image *gfi, int colormaps)
   if (gfi->transparent >= 0)
     fprintf(where, " transparent %d", gfi->transparent);
   
-#ifdef PRINT_SIZE
+#if defined(PRINT_SIZE)
   if (gfi->compressed)
-    fprintf(where, " compressed size %u min_bits %d", gfi->compressed_len,
-	    *gfi->compressed);
+    fprintf(where, " compressed size %u min_code_size %d", gfi->compressed_len, *gfi->compressed);
 #endif
   
   fprintf(where, "\n");
@@ -423,7 +439,7 @@ image_info(FILE *where, Gif_Stream *gfs, Gif_Image *gfi, int colormaps)
 
 
 char *
-explode_filename(char *filename, int number, char *name)
+explode_filename(char *filename, int number, char *name, int max_nimages)
 {
   static char *s;
   int l = strlen(filename);
@@ -433,8 +449,16 @@ explode_filename(char *filename, int number, char *name)
   s = Gif_NewArray(char, l + 3);
   if (name)
     sprintf(s, "%s.%s", filename, name);
-  else
-    sprintf(s, "%s.%d", filename, number);
+  else if (max_nimages <= 1000)
+    sprintf(s, "%s.%03d", filename, number);
+  else {
+    int digits;
+    unsigned j;
+    unsigned max = (max_nimages < 0 ? 0 : max_nimages);
+    for (digits = 4, j = 10000; max > j; digits++)
+      j *= 10;
+    sprintf(s, "%s.%0*d", filename, digits, number);
+  }
   
   return s;
 }
@@ -509,7 +533,11 @@ parse_frame_spec(Clp_Parser *clp, const char *arg, int complain, void *thunk)
       frame_spec_name = (char *)arg;
       frame_spec_1 = frame_spec_2 = Gif_ImageNumber(input, gfi);
       return 1;
-    } else if (complain)
+    } else if (complain < 0)	/* -1 is special value meaning `don't complain
+                                   about frame NAMES, but do complain about
+                                   frame numbers.' */
+      return -97;		/* Return -97 on bad frame name. */
+    else if (complain)
       return Clp_OptionError(clp, "no frame named `#%s'", arg);
     else
       return 0;
@@ -537,10 +565,18 @@ int
 parse_dimensions(Clp_Parser *clp, const char *arg, int complain, void *thunk)
 {
   char *val;
-  
-  dimensions_x = strtol(arg, &val, 10);
+
+  if (*arg == '_' && arg[1] == 'x') {
+    dimensions_x = 0;
+    val = (char *)(arg + 1);
+  } else
+    dimensions_x = strtol(arg, &val, 10);
   if (*val == 'x') {
-    dimensions_y = strtol(val + 1, &val, 10);
+    if (val[1] == '_' && val[2] == 0) {
+      dimensions_y = 0;
+      val = val + 2;
+    } else
+      dimensions_y = strtol(val + 1, &val, 10);
     if (*val == 0)
       return 1;
   }
@@ -958,10 +994,12 @@ merger_flatten(Gt_Frameset *fset, int f1, int f2)
     
     if (nest && nest->count > 0) {
       if (FRAME(fset, i).use < 0 && nest->count == 1) {
-	/* use < 0 means use the frame's delay and name (if not explicitly
-	   overridden), but not the frame itself. */
+	/* use < 0 means use the frame's delay, disposal and name (if not
+	   explicitly overridden), but not the frame itself. */
 	if (FRAME(nest, 0).delay < 0)
 	  FRAME(nest, 0).delay = FRAME(fset, i).image->delay;
+	if (FRAME(nest, 0).disposal < 0)
+	  FRAME(nest, 0).disposal = FRAME(fset, i).image->disposal;
 	if (FRAME(nest, 0).name == 0 && FRAME(nest, 0).no_name == 0)
 	  FRAME(nest, 0).name =
 	    Gif_CopyString(FRAME(fset, i).image->identifier);
@@ -991,8 +1029,8 @@ find_background(Gif_Colormap *dest_global, Gif_Color *background)
       static int context = 0;
       warning("irrelevant background color");
       if (!context) {
-	warning("(The background will appear transparent because");
-	warning("the first image contains transparency.)");
+	warncontext("(The background will appear transparent because");
+	warncontext("the first image contains transparency.)");
 	context = 1;
       }
     }
@@ -1042,7 +1080,7 @@ find_background(Gif_Colormap *dest_global, Gif_Color *background)
 	static int context = 0;
 	warning("input images have conflicting background colors");
 	if (!context) {
-	  warning("(This means some animation frames may appear incorrect.)");
+	  warncontext("(This means some animation frames may appear incorrect.)");
 	  context = 1;
 	}
 	break;
@@ -1295,9 +1333,9 @@ merge_frame_interval(Gt_Frameset *fset, int f1, int f2,
     if (fr->interlacing >= 0)
       desti->interlace = fr->interlacing;
     if (fr->left >= 0)
-      desti->left = fr->left;
+      desti->left = fr->left + (fr->position_is_offset ? desti->left : 0);
     if (fr->top >= 0)
-      desti->top = fr->top;
+      desti->top = fr->top + (fr->position_is_offset ? desti->top : 0);
     
     if (fr->delay >= 0)
       desti->delay = fr->delay;
